GLOBAL.setmetatable(env,{__index=function(t,k) return GLOBAL.rawget(GLOBAL,k) end})

local function DummyFn() end

-------------------------------------------------------------------------------------------------

-- Using ACTIONS.ATTACK handler instead of "doattack" event handler.

ACTIONS.ATTACK.instant  = false

ACTIONS.ATTACK.fn = function(act)
    if act.target.components.combat ~= nil then
        act.doer.components.combat:SetTarget(act.target)
        act.doer.components.combat:_DoAttack(act.target)

        return true
    end
end

-------------------------------------------------------------------------------------------------

-- Using ACTIONS.ATTACK handler instead of "doattack" event handler.

local AttackActionHandler = ActionHandler(ACTIONS.ATTACK, "attack_proxy")

AddStategraphActionHandler("wilson",        AttackActionHandler)
AddStategraphActionHandler("wilsonboating", AttackActionHandler)

-------------------------------------------------------------------------------------------------

local function GoToAttackState(inst)
    inst.sg:GoToState("attack")
end

local function GoToIdleState(inst)
    inst.sg:GoToState("idle")
end

local function WilsonStategraphPostInit(sg)
    local _locomote_handler = sg.events.locomote ~= nil and sg.events.locomote.fn or nil

    if (_locomote_handler or sg.states.attack) == nil then return end

    ---------------------------------------------------------------------------

    -- Using ACTIONS.ATTACK handler instead of "doattack" event handler, but still using the original function.

    local _doattack_handler = sg.events.doattack.fn or GoToAttackState

    sg.states.attack_proxy = State{
        name = "attack_proxy",
        onenter = _doattack_handler,
    }

    ---------------------------------------------------------------------------

    -- Hack to remove/ignore the "attack" state tag check in "locomote" event handler.

    sg.events.locomote.fn = function(inst, ...)
        local had_tag = inst.sg:HasStateTag("attack") or nil
        local state = inst.sg.currentstate

        inst.sg:RemoveStateTag("attack")

        _locomote_handler(inst, ...) -- This might change the state.

        if had_tag and state == inst.sg.currentstate then
            inst.sg:AddStateTag("attack")
        end
    end

    ---------------------------------------------------------------------------

    -- Some improvements to attack states...

    local _attack_onenter = sg.states.attack.onenter or DummyFn

    sg.states.attack.onenter = function(inst, ...)
        _attack_onenter(inst, ...)

        local timeout = inst.sg.timeout or ((inst.sg.statemem.slow or inst.sg.statemem.slowweapon) and 24 or 13) * FRAMES
        inst.sg:SetTimeout(math.max(timeout, inst.components.combat.min_attack_period))

        local buffaction = inst:GetBufferedAction()
        local target = buffaction ~= nil and buffaction.target or nil

        if target ~= nil then
            -- The chaseandattack behavior is no longer used, so call BattleCry here.
            inst.components.combat:BattleCry()
        end
    end

    -- NOTE: Overriding the timeline!
    sg.states.attack.timeline = {
        TimeEvent(8 * FRAMES, function(inst)
            inst:PerformBufferedAction() -- Important!
            inst.sg:RemoveStateTag("abouttoattack")

            local weapon = inst.components.combat:GetWeapon()

            if weapon ~= nil and weapon:HasTag("corkbat") then
                inst.SoundEmitter:PlaySound("dontstarve_DLC003/common/items/weapon/corkbat_hit")
            end
        end),
    }

    local _attack_ontimeout = sg.states.attack.ontimeout or DummyFn

    sg.states.attack.ontimeout = function(inst, ...)
        _attack_ontimeout(inst, ...)

        inst.sg:RemoveStateTag("attack")
        inst.sg:AddStateTag("idle")
    end

    sg.states.attack.events.equip   = EventHandler("equip",   GoToIdleState)
    sg.states.attack.events.unequip = EventHandler("unequip", GoToIdleState)

    ---------------------------------------------------------------------------

    for name, data in pairs(sg.states) do
        if data.tags.attack then
            -- Removing busy state tag from attack states.
            data.tags.busy = nil

            local _onenter = data.onenter or DummyFn

            data.onenter = function(inst, ...)
                if inst.components.combat:InCooldown() then
                    inst.sg:RemoveStateTag("abouttoattack")
                    inst:ClearBufferedAction()
                    inst.sg:GoToState("idle", true)
                    return
                end

                _onenter(inst, ...)
            end

            local _onexit = data.onexit or DummyFn

            data.onexit = function(inst, ...)
                _onexit(inst, ...)

                inst.components.combat:SetTarget(nil)

                if inst.sg:HasStateTag("abouttoattack") then
                    inst.components.combat.laststartattacktime = nil
                end
            end
        end
    end
end

AddStategraphPostInit("wilson",        WilsonStategraphPostInit)
AddStategraphPostInit("wilsonboating", WilsonStategraphPostInit)

-------------------------------------------------------------------------------------------------

AddPlayerPostInit(function(player)
    -- Removing chase and attack behavior...
    player:SetBrain()

    ---------------------------------------------------------------------------

    -- Hacky way to force combat:DoAttack into inst:PerformBufferedAction.
    -- Really problematic in several ways...

    local combat = player.components.combat

    if combat ~= nil then
        combat._DoAttack = combat.DoAttack

        if combat._DoAttack ~= nil then
            function combat:DoAttack(target_override, weapon, projectile, ...)
                if not (weapon or projectile) and self.inst.bufferedaction ~= nil and self.inst.sg ~= nil and self.inst.sg:HasStateTag("attack") then
                    self.inst:PerformBufferedAction()

                    return
                end

                return self:_DoAttack(target_override, weapon, projectile, ...)
            end
        end

        local _CanTarget = combat.CanTarget
        
        function combat:CanTarget(...)
            -- Can't target anything when attacking. This prevents left click spam.
            if self.inst.sg ~= nil and self.inst.sg:HasStateTag("attack") then
                return false
            end

            return _CanTarget(combat, ...)
        end

    end

    ---------------------------------------------------------------------------

    -- ACTIONS.ATTACK custom arride_dist.

    local locomotor = player.components.locomotor
    local _LocomotorOnUpdate = locomotor ~= nil and locomotor.OnUpdate or nil

    if _LocomotorOnUpdate == nil then return end

    function locomotor:OnUpdate(...)
        if not self.inst:IsValid() then
            Print(VERBOSITY.DEBUG, "OnUpdate INVALID", self.inst.prefab)
            self:ResetPath()
            self.inst:StopUpdatingComponent(self)
            return
        end

        local _arrive_dist = self.arrive_dist

        if self.dest then
            if not self.dest:IsValid() or (self.bufferedaction and not self.bufferedaction:IsValid()) then
                self:Clear()
                return
            end

            if self.inst.components.health and self.inst.components.health:IsDead() then
                self:Clear()
                return
            end

            -- New code: Custom check for ACTIONS.ATTACK, from DST.

            if self.bufferedaction ~= nil and
                self.bufferedaction.action == ACTIONS.ATTACK and
                self.bufferedaction.target ~= nil and
                self.bufferedaction.target:IsValid() and
                self.inst.components.combat ~= nil
            then
                local can_hit_target = self.inst.components.combat:CanHitTarget(self.bufferedaction.target, self.inst.components.combat:GetWeapon())

                if can_hit_target then
                    self.arrive_dist = math.huge -- Forcing the bufferedaction push.
                end
            end
        end

        _LocomotorOnUpdate(self, ...)

        self.arrive_dist = _arrive_dist
    end
end)

---------------------------------------------------------------------------

-- Don't find a new target until we can actually attack. This makes the timing of the states correct.

AddComponentPostInit("playercontroller", function(self)
    local _GetAttackTarget = self.GetAttackTarget or DummyFn

    function self:GetAttackTarget(...)
        if self.inst.sg ~= nil and self.inst.sg:HasStateTag("attack") then
            return
        end

        return _GetAttackTarget(self, ...)
    end

    -- Overriding it.
    function self:DoAttackButton()
        local taregt = self:GetAttackTarget(TheInput:IsControlPressed(CONTROL_FORCE_ATTACK))
    
        if taregt ~= nil then
            local action = BufferedAction(self.inst, taregt, ACTIONS.ATTACK)
            self.inst.components.locomotor:PushAction(action, true)
        end
    end
end)